#include <linux/blkdev.h>
#include <linux/clk.h>
#include <linux/debugfs.h>
#include <linux/device.h>
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include <linux/err.h>
#include <linux/gpio.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/scatterlist.h>
#include <linux/seq_file.h>
#include <linux/stat.h>

#include <linux/mmc/host.h>

#include <asm/io.h>
#include <asm/unaligned.h>

#include <mach/hardware.h>
#include <mach/regs-clock.h>
#include <mach/regs-irq.h>

//#define NUC900_SD_DEBUG
#undef NUC900_SD_DEBUG

#ifdef NUC900_SD_DEBUG
#define PDEBUG(fmt, arg...)		printk(fmt, ##arg)
#else
#define PDEBUG(fmt, arg...)
#endif

#define REG_MFSEL	(W90X900_VA_GCR + 0xC)

/* DMAC Control Registers */
#define NUC900_DMAFB		0x000  /* DMAC Frame buffer */
#define NUC900_DMACCSR		0x800  /* DMAC Control and Status Register */
#define NUC900_DMACSAR1		0x804  /* DMAC Transfer Starting Address Register 1 */
#define NUC900_DMACSAR2		0x808  /* DMAC Transfer Starting Address Register 2 */
#define NUC900_DMACBCR		0x80C  /* DMAC Transfer Byte Count Register */
#define NUC900_DMACIER		0x810  /* DMAC Interrupt Enable Register */
#define NUC900_DMACISR		0x814  /* DMAC Interrupt Status Register */

/* Flash Memory Card Interface Registers */
#define NUC900_FMICSR		0x000  /* Global Control and Status Register */
#define NUC900_FMIIER		0x004  /* Global Interrupt Control Register */
#define NUC900_FMIISR		0x008  /* Global Interrupt Status Register */

/* Secure Digit Registers */
#define NUC900_SDCSR		0x020  /* SD control and status register */
#define NUC900_SDARG		0x024  /* SD command argument register */
#define NUC900_SDIER		0x028  /* SD interrupt enable register */
#define NUC900_SDISR		0x02C  /* SD interrupt status register */
#define NUC900_SDRSP0		0x030  /* SD receive response token register 0 */
#define NUC900_SDRSP1		0x034  /* SD receive response token register 1 */
#define NUC900_SDBLEN		0x038  /* SD block length register */

/* DMAC Control and Status Register (DMACCSR) */
#define DMACCSR_DMACEN		(1<<0)
#define DMACCSR_SW_RST		(1<<1)
#define DMACCSR_SG_EN1		(1<<2)
#define DMACCSR_SG_EN2		(1<<3)
#define DMACCSR_ATA_BUSY	(1<<8)
#define DMACCSR_FMI_BUSY	(1<<9)

/* DMAC Interrupt Enable Register (DMACIER) */
#define DMACIER_TABORT_IE	(1<<0)
#define DMACIER_WEOT_IE		(1<<1)

/* DMAC Interrupt Status Register (DMACISR) */
#define DMACISR_TABORT_IF	(1<<0)
#define DMACISR_WEOT_IF		(1<<1)

/* DMAC BIST Control and Status Register (DMACBIST) */
#define DMACBIST_BIST_EN	(1<<0)
#define DMACBIST_FINISH		(1<<1)
#define DMACBIST_FAILED		(1<<2)

/* FMI Global Control and Status Register(FMICSR) */
#define FMICSR_SW_RST       (1<<0)
#define FMICSR_SD_EN        (1<<1)
#define FMICSR_MS_EN        (1<<2)
#define FMICSR_SM_EN        (1<<3)
#define FMICSR_CF_EN        (1<<4)

/* FMI Global Interrupt Control Register(FMIIER) */
#define FMIIER_DTA_IE       (1<<0)

/* FMI Global Interrupt Status Register (FMIISR) */
#define FMIISR_DTA_IF       (1<<0)

/* SD Control and Status Register (SDCSR) */
#define SDCSR_CO_EN          (1<<0)
#define SDCSR_RI_EN          (1<<1)
#define SDCSR_DI_EN          (1<<2)
#define SDCSR_DO_EN          (1<<3)
#define SDCSR_R2_EN          (1<<4)
#define SDCSR_CLK74_OE       (1<<5)
#define SDCSR_CLK8_OE        (1<<6)
#define SDCSR_CLK_KEEP0      (1<<7)
#define SDCSR_SW_RST         (1<<14)
#define SDCSR_DBW            (1<<15)
#define SDCSR_CLK_KEEP1      (1<<31)

#define SDCSR_CMD_CODE_SHIFT 8
#define SDCSR_CMD_CODE_MASK  (0x3F<<SDCSR_CMD_CODE_SHIFT)
#define SDCSR_BLK_CNT_SHIFT  16
#define SDCSR_BLK_CNT_MASK   (0xFF<<SDCSR_BLK_CNT_SHIFT)
#define SDCSR_SDNWR_SHIFT    24
#define SDCSR_SDNWR_MASK     (0x0F<<SDCSR_SDNWR_SHIFT)
#define SDCSR_SDPORT_SHIFT   29
#define SDCSR_SDPORT_MASK    (0x03<<SDCSR_SDPORT_SHIFT)

/* SD Interrupt Control Register (SDIER) */
#define SDIER_BLKD_IE       (1<<0)
#define SDIER_CRC_IE        (1<<1)
#define SDIER_CD0_IE        (1<<8)
#define SDIER_CD1_IE        (1<<9)
#define SDIER_SDIO0_IE      (1<<10)
#define SDIER_SDIO1_IE      (1<<11)
#define SDIER_WKUP_EN       (1<<12)
#define SDIER_CD0SRC        (1<<30)
#define SDIER_CD1SRC        (1<<31)

/* SD Interrupt Status Register (SDISR) */
#define SDISR_BLKD_IF       (1<<0)
#define SDISR_CRC_IF        (1<<1)
#define SDISR_CRC_7         (1<<2)
#define SDISR_CRC_16        (1<<3)
#define SDISR_SDDAT0        (1<<7)
#define SDISR_CD0_IF        (1<<8)
#define SDISR_CD1_IF        (1<<9)
#define SDISR_SDIO0_IF      (1<<10)
#define SDISR_SDIO1_IF      (1<<11)
#define SDISR_CDPS0         (1<<16)
#define SDISR_CDPS1         (1<<17)
#define SDISR_SD0DAT1       (1<<18)
#define SDISR_SD1DAT1       (1<<19)


#define nuc900_sdio_read(host, reg)	__raw_readl((host)->baseaddr + (reg))
#define nuc900_sdio_write(host, reg, val)	__raw_writel((val), (host)->baseaddr + (reg))
#define nuc900_sdio_wrset(host, reg, val) do \
	{ \
		register unsigned int tmp_val; \
		tmp_val  = nuc900_sdio_read(host, reg); \
		tmp_val |= val; \
		nuc900_sdio_write(host, reg, tmp_val); \
	} while(0);
#define nuc900_sdio_wrclr(host, reg, val) do \
	{ \
		register unsigned int tmp_val; \
		tmp_val  = nuc900_sdio_read(host, reg); \
		tmp_val &= (~val); \
		nuc900_sdio_write(host, reg, tmp_val); \
	} while(0);
	
#define nuc900_dmac_read(host, reg) __raw_readl((host)->dma_baseaddr + (reg))
#define nuc900_dmac_write(host, reg, val)	__raw_writel((val), (host)->dma_baseaddr + (reg))
#define nuc900_dmac_wrset(host, reg, val) do \
	{ \
		register unsigned int tmp_val; \
		tmp_val  = nuc900_dmac_read(host, reg); \
		tmp_val |= val; \
		nuc900_dmac_write(host, reg, tmp_val); \
	} while(0);
#define nuc900_dmac_wrclr(host, reg, val) do \
	{ \
		register unsigned int tmp_val; \
		tmp_val  = nuc900_dmac_read(host, reg); \
		tmp_val &= (~val); \
		nuc900_dmac_write(host, reg, tmp_val); \
	} while(0);
	
#define Enable_IRQ(n)     				__raw_writel(1 << (n),REG_AIC_MECR)
#define Disable_IRQ(n)    				__raw_writel(1 << (n),REG_AIC_MDCR)

#define FMI_INPUT_CLOCK 200000000

#define NUC_DATA_NONE 0
#define NUC_DATA_READ 1
#define NUC_DATA_WRITE 2

struct nuc900_pad_table
{
	unsigned long pa;
	unsigned long num   : 16;
	unsigned long dummy : 15;
	unsigned long eot   :  1;
};

/*
 * Low level type for this driver
 */
struct nuc900_sdio_host
{
	struct mmc_host *mmc;
	struct mmc_command *cmd;
	struct mmc_request *request;

	void __iomem *baseaddr;
	void __iomem *dma_baseaddr;
	int irq;

	int present;

	struct clk *mci_clk;

	/*
	 * Flag indicating when the command has been sent. This is used to
	 * work out whether or not to send the stop
	 */
	unsigned int flags;
	/* flag for current bus settings */
	u32 bus_mode;

	/* DMA buffer used for transmitting */
	unsigned int* buffer;
	dma_addr_t physical_address;
	unsigned int total_length;

	/* Latest in the scatterlist that has been enabled for transfer, but not freed */
	int in_use_index;

	/* Latest in the scatterlist that has been enabled for transfer */
	int transfer_index;

	/* Timer for timeouts */
	struct timer_list timer;
};

/* Handle an interrupt */
static irqreturn_t nuc900_sdio_irq(int irq, void *devid)
{
	struct nuc900_sdio_host *host = devid;
	unsigned int status;

	status = nuc900_sdio_read(host, NUC900_SDISR);
	PDEBUG("SD int: %08lX\n", status);

	if (status & SDISR_BLKD_IF) // block transfer done
	{
		nuc900_sdio_write(host, NUC900_SDISR, SDISR_BLKD_IF);
	}

	if(status & SDISR_CD0_IF) //sd0 card inserted or removed
	{
		if (status & SDISR_CDPS0)
		{
			host->present = 0;
			PDEBUG("sd out\n");
		}
		else
		{
			host->present = 1;
			PDEBUG("sd in\n");
		}
		mmc_detect_change(host->mmc, msecs_to_jiffies(200));
		nuc900_sdio_write(host, NUC900_SDISR, SDISR_CD0_IF);
	}

	return IRQ_HANDLED;
}

unsigned int nuc900_sdio_cmd(struct mmc_host *mmc, struct mmc_command *cmd, unsigned int dat)
{
	struct nuc900_sdio_host *host = mmc_priv(mmc);
	unsigned int sdcsr_tmp;
	unsigned int wait_mask;
	unsigned long orig_jiffies;
	unsigned long timeout = HZ / 100;

	sdcsr_tmp = nuc900_sdio_read(host, NUC900_SDCSR) & 0xFFFFC080;
	if (mmc_resp_type(cmd) != MMC_RSP_NONE) // need a response
	{
		if (mmc_resp_type(cmd) == MMC_RSP_R2)
		{
			PDEBUG("mmc need R2 response\n");
			sdcsr_tmp |= SDCSR_R2_EN; // host will wait a R2 response
			wait_mask = SDCSR_R2_EN;
		}
		else
		{
			PDEBUG("mmc need response\n");
			sdcsr_tmp |= SDCSR_RI_EN; // host will wait a response
			wait_mask = SDCSR_RI_EN;
		}
	}
	else
		wait_mask = SDCSR_CO_EN;
		
	if (dat == NUC_DATA_READ)
	{
		sdcsr_tmp |= SDCSR_DI_EN;
		wait_mask = SDCSR_DI_EN;
		timeout = HZ / 1;
	}
	else if (dat == NUC_DATA_WRITE)
	{
		sdcsr_tmp |= SDCSR_DO_EN;
		wait_mask = SDCSR_DO_EN;
		timeout = HZ / 1;
	}

	sdcsr_tmp |= SDCSR_CO_EN | (cmd->opcode<<SDCSR_CMD_CODE_SHIFT);
	PDEBUG("write SDCSR %08lX\n", sdcsr_tmp);

	nuc900_sdio_write(host, NUC900_SDARG, cmd->arg);
	nuc900_sdio_write(host, NUC900_SDCSR, sdcsr_tmp);

	orig_jiffies = jiffies;
	while(1)
	{
		if ( !(nuc900_sdio_read(host, NUC900_SDCSR) & wait_mask) )
			break;
		if (time_after(jiffies, orig_jiffies + timeout))
		{
			nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_SW_RST); // reset SDIO
			PDEBUG("mmc command timeout\n");
			cmd->error = -ETIMEDOUT;
			return -ETIMEDOUT;
		}
		schedule();
	}

	if (wait_mask == SDCSR_R2_EN)
	{
		unsigned char *resp_ptr = (unsigned char*)&cmd->resp[0];
		unsigned long i, a;

		for(i=0;i<16;i++)
		{
			a <<= 8;
			a  |= __raw_readb(host->dma_baseaddr + NUC900_DMAFB + 1 + i);
			if ((i&3) == 3)
				cmd->resp[i>>2] = a;
		}
		PDEBUG("mmc R2 responce: %08lX-%08lX-%08lX-%08lX\n", cmd->resp[0], cmd->resp[1], cmd->resp[2], cmd->resp[3]);
	}
	else
	if (wait_mask == SDCSR_RI_EN)
	{
		unsigned long resp_hi, resp_lo;

		resp_hi = nuc900_sdio_read(host, NUC900_SDRSP0);
		resp_lo = nuc900_sdio_read(host, NUC900_SDRSP1);
		cmd->resp[0] = (resp_hi<<8) | resp_lo;
		PDEBUG("mmc responce: %08lX\n", cmd->resp[0]);
	}

	return 0;
}

/* Handle an MMC request */
static void nuc900_sdio_request(struct mmc_host *mmc, struct mmc_request *request)
{
	struct nuc900_sdio_host *host = mmc_priv(mmc);
	struct mmc_command *cmd = request->cmd;
	struct mmc_command *stop = request->stop;
	struct mmc_data *data = request->data;
	struct nuc900_pad_table *pad_table;
	unsigned int dma_nents;
	unsigned int dma_size;
	unsigned int dma_dir;
	unsigned int ret;
	unsigned int tmp;
	unsigned int timeout = 0;
	unsigned int dat_dir = NUC_DATA_NONE;

	PDEBUG("------- SD REQUEST -------\n");

	host->request = request;

	PDEBUG("mmc request: cmd=%u, arg=%08X, flags=%08lX\n", cmd->opcode, cmd->arg, cmd->flags);
	if (data) {
		unsigned int i;

		PDEBUG("with data: ");
		if (data->flags & MMC_DATA_WRITE) {
			dma_dir = DMA_TO_DEVICE;
			PDEBUG("write, ");
			dat_dir = NUC_DATA_WRITE;
		}
		else {
			dma_dir = DMA_FROM_DEVICE;
			PDEBUG("read, ");
			dat_dir = NUC_DATA_READ;
		}
		if (data->flags & MMC_DATA_STREAM)
			PDEBUG("stream, ");
		PDEBUG("block size=%u, number=%lu\n", data->blksz, data->blocks);

		PDEBUG("DMA sg:\n");
		dma_nents = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,  dma_dir);
		for(i=0;i<dma_nents;i++)
		{
			PDEBUG("addr=0x%08lX, len=%lu\n", data->sg[i].dma_address, data->sg[i].length);
		}

		// enable and reset DMA
		//nuc900_dmac_wrset(host, NUC900_DMACCSR, DMACCSR_SW_RST);

		if (data->blksz < 512)
			nuc900_sdio_write(host, NUC900_SDBLEN, data->blksz-1);
		else
			nuc900_sdio_write(host, NUC900_SDBLEN, 0x1FF);


		// disable scatter-gather mode, i.e. set normal mode
		nuc900_dmac_wrclr(host, NUC900_DMACCSR, DMACCSR_SG_EN2);
		// write start address of data memory for DMA
		nuc900_dmac_write(host, NUC900_DMACSAR2, data->sg[0].dma_address);
		// write number of blocks
		tmp  = nuc900_sdio_read(host, NUC900_SDCSR) & (~SDCSR_BLK_CNT_MASK);
		tmp |= ( (data->sg[0].length/512) << SDCSR_BLK_CNT_SHIFT );
		PDEBUG("Write SDCSR %08lX\n", tmp);
		nuc900_sdio_write(host, NUC900_SDCSR, tmp);

		// enable DMA
		//nuc900_dmac_wrset(host, NUC900_DMACCSR, DMACCSR_DMACEN);
	}
	ret = nuc900_dmac_read(host, NUC900_DMACCSR);
	PDEBUG("DMACCSR=%08lX\n", ret);

	nuc900_sdio_cmd(mmc, cmd, dat_dir);

	if (data) {
		unsigned int i;
		unsigned int wait_mask;
		unsigned long orig_jiffies;
		
		for(i=1;i<dma_nents;i++) {
			if (cmd->error == -ETIMEDOUT)
				break;
			// write start address of data memory for DMA
			nuc900_dmac_write(host, NUC900_DMACSAR2, data->sg[i].dma_address);
			// write number of blocks
			tmp  = nuc900_sdio_read(host, NUC900_SDCSR) & (~SDCSR_BLK_CNT_MASK);
			tmp |= ( (data->sg[i].length/512) << SDCSR_BLK_CNT_SHIFT );
			if (dat_dir == NUC_DATA_WRITE)
				wait_mask = SDCSR_DO_EN;
			else
				wait_mask = SDCSR_DI_EN;
			tmp |= wait_mask; // to start transfer
			PDEBUG("Write SDCSR %08lX\n", tmp);
			nuc900_sdio_write(host, NUC900_SDCSR, tmp);

			orig_jiffies = jiffies;
			while(1)
			{
				if ( !(nuc900_sdio_read(host, NUC900_SDCSR) & wait_mask) )
					break;
				if (time_after(jiffies, orig_jiffies + (HZ / 1)))
				{
					nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_SW_RST); // reset SDIO
					PDEBUG("mmc command timeout\n");
					cmd->error = -ETIMEDOUT;
					break;
				}
				schedule();
			}
		}

		PDEBUG("sd DMA complited\n");
		dma_unmap_sg(mmc_dev(host->mmc), data->sg, dma_nents, dma_dir);
		data->bytes_xfered = data->blksz * data->blocks;
		data->error = 0;
	}

	host->cmd = NULL;

	if (host->request->stop) {
		nuc900_sdio_cmd(mmc, stop, NUC_DATA_NONE);
	}

	mmc_request_done(host->mmc, host->request);
}

/* Set SD controller clock  */
static void nuc900_sdio_set_clock(struct nuc900_sdio_host *host, unsigned int clock)
{
	unsigned int rate, tmp;
	
	if (clock == 0)
	{
		return;
	}
	
	rate = (FMI_INPUT_CLOCK) / clock;
	if (rate > 0xFF) // 0xFF is a maximum divider
		rate = 0xFF;
	tmp = __raw_readl(REG_CLKDIV1);
	tmp = (tmp & 0xFFFF00FF) | (rate << 8); // REG_CLKDIV1[15..8] - SD divider
	__raw_writel(tmp, REG_CLKDIV1);
}

/* Set the IOS */
static void nuc900_sdio_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct nuc900_sdio_host *host = mmc_priv(mmc);
	unsigned int sdcsr_tmp;
	
	PDEBUG("------- SD IOS IN -------\n");

	if (ios->bus_width == MMC_BUS_WIDTH_4) {
		PDEBUG("mmc set width 4\n");
		nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_DBW);
	}
	else {
		PDEBUG("mmc set width 1\n");
		nuc900_sdio_wrclr(host, NUC900_SDCSR, SDCSR_DBW);
	}
	
	if (ios->clock) {
		PDEBUG("mmc set clock to %lu Hz\n", ios->clock);
		nuc900_sdio_set_clock(host, ios->clock);
	}
	else {
	
	}	

	switch (ios->power_mode) {
	case MMC_POWER_ON:
		PDEBUG("mmc power on\n");
		
		sdcsr_tmp  = nuc900_sdio_read(host, NUC900_SDCSR) & 0xFFFFC080;
		sdcsr_tmp |= SDCSR_CLK74_OE;
		nuc900_sdio_write(host, NUC900_SDCSR, sdcsr_tmp);
		while( nuc900_sdio_read(host, NUC900_SDCSR) &  SDCSR_CLK74_OE )
		{
			schedule();
		}
		
		//nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_CLK_KEEP0);
		PDEBUG("power on ok!\n");
		break;
	case MMC_POWER_UP:
		nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_SW_RST); // reset SDIO
		PDEBUG("mmc power up\n");
		break;
	case MMC_POWER_OFF:
		nuc900_sdio_wrclr(host, NUC900_SDCSR, SDCSR_CLK_KEEP0);
		PDEBUG("mmc power off\n");
		break;
	default:
		PDEBUG("no!\n");
	}
	PDEBUG("------- SD IOS OUT -------\n");
}

static int nuc900_sdio_get_ro(struct mmc_host *mmc)
{
	/*
	 * Board doesn't support read only detection; let the mmc core
	 * decide what to do.
	 */
	return -ENOSYS;
}

static int nuc900_sdio_card_present(struct mmc_host *mmc)
{
	struct nuc900_sdio_host *host = mmc_priv(mmc);
	unsigned int ret;
	
	ret = nuc900_sdio_read(host, NUC900_SDISR) & SDISR_CDPS0;
	
	if (!ret)
	{
		PDEBUG("mmc in\n");
		return 1;
	}
	else
	{
		PDEBUG("mmc out\n");
		return 0;
	}
}

static void nuc900_sdio_enable_irq(struct mmc_host *mmc, int enable)
{
	if (enable)
		PDEBUG("mmc enable irq\n");
	else
		PDEBUG("mmc disable irq\n");
}

static const struct mmc_host_ops nuc900_sdio_ops = {
	.request	= nuc900_sdio_request,
	.set_ios	= nuc900_sdio_set_ios,
	.get_ro		= nuc900_sdio_get_ro,
	.get_cd		= nuc900_sdio_card_present,
	.enable_sdio_irq = nuc900_sdio_enable_irq,
};

static int __devinit nuc900_sdio_probe(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct nuc900_sdio_host *host;
	struct resource *res;
	int ret;

	PDEBUG("NUC900 SDIO probe\n");
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENXIO;

	if (!request_mem_region(res->start, res->end - res->start + 1, "nuc900-fmi"))
		return -EBUSY;

	mmc = mmc_alloc_host(sizeof(struct nuc900_sdio_host), &pdev->dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto fail6;
	}

	mmc->ops = &nuc900_sdio_ops;
	mmc->f_min = 350000;
	mmc->f_max = 25000000;
	mmc->ocr_avail = MMC_VDD_32_33 | MMC_VDD_33_34;
	mmc->caps = MMC_CAP_SDIO_IRQ | MMC_CAP_4_BIT_DATA;

	// MMC core transfer sizes tunable parameters
	mmc->max_hw_segs = 64;
	mmc->max_phys_segs = 64;
	mmc->max_seg_size = 64*512; // default PAGE_CACHE_SIZE
	mmc->max_req_size = 64*512; // default PAGE_CACHE_SIZE
	mmc->max_blk_size = 512;
	mmc->max_blk_count = 65535;

	host = mmc_priv(mmc);
	host->mmc = mmc;
	host->buffer = NULL;
	host->bus_mode = 0;

	// Map I/O region
	host->baseaddr = ioremap(res->start, res->end - res->start + 1);
	if (!host->baseaddr) {
		ret = -ENOMEM;
		goto fail1;
	}

	// Map DMAC I/O region
	host->dma_baseaddr = ioremap(W90X900_PA_DMA, 0x1000);
	if (!host->baseaddr) {
		ret = -ENOMEM;
		goto fail1;
	}

	// Reset hardware
	// enable DMAC and FMI clock
	__raw_writel(__raw_readl(REG_CLKEN) | 0x30,REG_CLKEN);
	// enable SD clock
	__raw_writel(__raw_readl(REG_CLKEN1) | 0x2,REG_CLKEN1);

	// reset FMI and enable SDIO functionality
	nuc900_sdio_wrset(host, NUC900_FMICSR, (FMICSR_SD_EN | FMICSR_SW_RST));
	// reset SDIO
	nuc900_sdio_wrset(host, NUC900_SDCSR, SDCSR_SW_RST);
	
	nuc900_sdio_write(host, NUC900_SDIER, SDIER_BLKD_IE | // block transfer done interrupt
	                                      SDIER_CD0_IE  | // card detection interrupt enable
	                                      SDIER_CD0SRC);  // from GPIO
	//DMACCSR_DMACEN
	nuc900_dmac_write(host, NUC900_DMACCSR, DMACCSR_DMACEN);
	ret = nuc900_dmac_read(host, NUC900_DMACCSR);
	PDEBUG("SD init DMACCSR=%08lX\n", ret);
	//SD 0 GPIO select
	__raw_writel((__raw_readl(REG_MFSEL)&0xFFFFFF0F)|0xa0,REG_MFSEL); //GPSELD[5:4]=10 ; [7:6]=10
	//__raw_writel(0x100,REG_GPIOD_DIR);	// set gpiod-8 to output
	//__raw_writel(__raw_readl(REG_GPIOD_DATAOUT) | 0x100,REG_GPIOD_DATAOUT);	// set gpiod-8 output high
	//udelay(1000);

	// Allocate the MCI interrupt
	host->irq = platform_get_irq(pdev, 0);
	ret = request_irq(host->irq, nuc900_sdio_irq, IRQF_SHARED,
	                  mmc_hostname(mmc), host);
	if (ret) {
		PDEBUG("request FMI(SDIO) interrupt failed\n");
		goto fail0;
	}
	
	Enable_IRQ(host->irq);

	platform_set_drvdata(pdev, mmc);

	// Add host to MMC layer
	host->present = -1;
	mmc_add_host(mmc);

	PDEBUG("OK NUC900 SDIO probe\n");
	
	return 0;
	
fail0:
	//clk_disable(host->mci_clk);
	iounmap(host->baseaddr);
fail1:
	//clk_put(host->mci_clk);
	mmc_free_host(mmc);
fail6:
	release_mem_region(res->start, res->end - res->start + 1);
	dev_err(&pdev->dev, "probe failed, err %d\n", ret);

	PDEBUG("Failed NUC900 SDIO probe\n");

	return ret;
}

static int __devexit nuc900_sdio_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver nuc900_sdio_driver = {
	.probe		= nuc900_sdio_probe,
	.remove		= __devexit_p(nuc900_sdio_remove),
	.driver		= {
		.name	= "nuc900-fmi",
		.owner	= THIS_MODULE,
	},
};

static int __init nuc900_sdio_init(void)
{
	return platform_driver_register(&nuc900_sdio_driver);
}

static void __exit nuc900_sdio_exit(void)
{
	platform_driver_unregister(&nuc900_sdio_driver);
}

module_init(nuc900_sdio_init);
module_exit(nuc900_sdio_exit);

MODULE_AUTHOR("I");
MODULE_DESCRIPTION("w90p950 sdio driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:w90p910-fmi");
